@@ -52,6 +52,9 @@ class OutputFormatter(object):
 
     progress = property(lambda self: self.options.progress)
     verbose = property(lambda self: self.options.verbose)
+    in_subprocess = property(
+        lambda self: self.options.resume_layer is not None and
+                     self.options.processes > 1)
 
     def compute_max_width(self):
         """Try to determine the terminal width."""
@@ -264,6 +267,12 @@ class OutputFormatter(object):
         elif self.verbose == 1:
             sys.stdout.write('.' * test.countTestCases())
 
+        elif self.in_subprocess:
+            sys.stdout.write('.' * test.countTestCases())
+            # Give the parent process a new line so it sees the progress
+            # in a timely manner.
+            sys.stdout.write('\n')
+
         if self.verbose > 1:
             s = str(test)
             sys.stdout.write(' ')
@@ -584,12 +593,15 @@ class ColorfulOutputFormatter(OutputFormatter):
                         self.color('normal'), '\n'])
                 else:
                     print line
-            elif line.startswith('    '):
+            elif line.startswith('    ') or line.strip() == '':
                 if colorize_diff and len(line) > 4:
                     color = self.diff_color.get(line[4], color_of_indented_text)
                     print self.colorize(color, line)
                 else:
-                    print self.colorize(color_of_indented_text, line)
+                    if line.strip() != '':
+                        print self.colorize(color_of_indented_text, line)
+                    else:
+                        print line
             else:
                 colorize_diff = False
                 if line.startswith('Failed example'):
